from __future__ import annotations

from typing import override

from mpris_server import (
  DbusObj, LoopStatus, MIME_TYPES, Metadata, Microseconds, MprisAdapter, PlayState,
  PlayerAdapter, Rate, RootAdapter, Track, TrackListAdapter, URI, Volume,
)

from .base import Device
from .device.wrapper import DeviceWrapper
from .protocols import DeviceIntegration


class DeviceRootAdapter(DeviceIntegration, RootAdapter):
  @override
  def can_quit(self) -> bool:
    return self.wrapper.can_quit()

  @override
  def get_desktop_entry(self) -> str:
    return self.wrapper.get_desktop_entry()

  @override
  def get_mime_types(self) -> list[str]:
    return MIME_TYPES

  @override
  def get_uri_schemes(self) -> list[str]:
    return URI

  @override
  def has_tracklist(self) -> bool:
    return self.wrapper.has_tracklist()

  @override
  def quit(self):
    self.wrapper.quit()


class DevicePlayerAdapter(DeviceIntegration, PlayerAdapter):
  @override
  def can_control(self) -> bool:
    return self.wrapper.can_control()

  @override
  def can_go_next(self) -> bool:
    return self.wrapper.can_play_next()

  @override
  def can_go_previous(self) -> bool:
    return self.wrapper.can_play_prev()

  @override
  def can_pause(self) -> bool:
    return self.wrapper.can_pause()

  @override
  def can_play(self) -> bool:
    return self.wrapper.can_play()

  @override
  def can_seek(self) -> bool:
    return self.wrapper.can_seek()

  @override
  def get_art_url(self, track: int = None) -> str:
    return self.wrapper.get_art_url(track)

  @override
  def get_current_position(self) -> Microseconds:
    return self.wrapper.get_current_position()

  @override
  def get_current_track(self) -> Track:
    return self.wrapper.get_current_track()

  @override
  def get_next_track(self) -> Track:
    return self.wrapper.get_next_track()

  @override
  def get_playstate(self) -> PlayState:
    return self.wrapper.get_playstate()

  @override
  def get_previous_track(self) -> Track:
    return self.wrapper.get_previous_track()

  @override
  def get_rate(self) -> Rate:
    return self.wrapper.get_rate()

  @override
  def get_shuffle(self) -> bool:
    return False

  @override
  def get_stream_title(self) -> str:
    return self.wrapper.get_stream_title()

  @override
  def get_volume(self) -> Volume:
    return self.wrapper.get_volume()

  @override
  def is_mute(self) -> bool:
    return self.wrapper.is_mute()

  @override
  def is_playlist(self) -> bool:
    return self.wrapper.is_playlist()

  @override
  def is_repeating(self) -> bool:
    return self.wrapper.is_repeating()

  @override
  def metadata(self) -> Metadata:
    return self.wrapper.metadata()

  @override
  def next(self):
    self.wrapper.next()

  @override
  def open_uri(self, uri: str):
    self.wrapper.open_uri(uri)

  @override
  def pause(self):
    self.wrapper.pause()

  @override
  def play(self):
    self.wrapper.play()

  @override
  def previous(self):
    self.wrapper.previous()

  @override
  def resume(self):
    self.play()

  @override
  def seek(self, time: Microseconds, track_id: DbusObj | None = None):
    self.wrapper.seek(time)

  @override
  def set_icon(self, lighter: bool = False):
    self.wrapper.set_icon(lighter)

  @override
  def set_loop_status(self, value: LoopStatus):
    pass

  @override
  def set_mute(self, value: bool):
    self.wrapper.set_mute(value)

  @override
  def set_rate(self, value: Rate):
    pass

  @override
  def set_repeating(self, value: bool):
    pass

  @override
  def set_shuffle(self, value: bool):
    pass

  @override
  def set_volume(self, value: Volume):
    self.wrapper.set_volume(value)

  @override
  def stop(self):
    self.wrapper.stop()


class DeviceTrackListAdapter(DeviceIntegration, TrackListAdapter):
  @override
  def add_track(self, uri: str, after_track: DbusObj, set_as_current: bool):
    self.wrapper.add_track(uri, after_track, set_as_current)

  @override
  def can_edit_tracks(self) -> bool:
    return self.wrapper.can_edit_tracks()

  @override
  def get_tracks(self) -> list[DbusObj]:
    return self.wrapper.get_tracks()


class DeviceAdapter(MprisAdapter, DevicePlayerAdapter, DeviceRootAdapter, DeviceTrackListAdapter):
  @override
  def __init__(self, device: Device):
    self.wrapper = DeviceWrapper(device)
    super().__init__(self.wrapper.name)



from __future__ import annotations

import logging
from time import sleep
from typing import Final, NoReturn
from uuid import UUID

from mpris_server import Server

from .daemon import Args, get_name
from .state import setup_logging
from ..adapter import DeviceAdapter
from ..base import DEFAULT_ICON, DEFAULT_RETRY_WAIT, DEFAULT_SET_LOG, DEFAULT_WAIT, LOG_LEVEL, \
  NoDevicesFound, Rc, Seconds
from ..device.device import find_device
from ..device.listeners import EventListener


log: Final[logging.Logger] = logging.getLogger(__name__)


def create_server(
  name: str | None = None,
  host: str | None = None,
  uuid: UUID | str | None = None,
  retry_wait: Seconds | None = DEFAULT_RETRY_WAIT,
) -> Server | None:
  if not (device := find_device(name, host, uuid, retry_wait)):
    return None

  adapter = DeviceAdapter(device)
  server = Server(name, adapter)

  EventListener.register(server, device)
  server.publish()

  return server


def retry_until_found(
  name: str | None = None,
  host: str | None = None,
  uuid: UUID | str | None = None,
  wait: Seconds | None = DEFAULT_WAIT,
  retry_wait: Seconds | None = DEFAULT_RETRY_WAIT,
) -> Server | None | NoReturn:
  """
    If the device isn't found, keep trying to find it.

    If `wait` is None, then retrying is disabled.
  """

  while True:
    if server := create_server(name, host, uuid, retry_wait):
      return server

    elif wait is None:
      return None

    device = get_name(name, host, uuid)
    log.warning(f'{device} not found. Waiting {wait} seconds before retrying.')
    sleep(wait)


def run_server(
  name: str | None = None,
  host: str | None = None,
  uuid: UUID | str | None = None,
  wait: Seconds | None = DEFAULT_WAIT,
  retry_wait: Seconds | None = DEFAULT_RETRY_WAIT,
  icon: bool = DEFAULT_ICON,
  log_level: str = LOG_LEVEL,
  set_logging: bool = DEFAULT_SET_LOG,
  background: bool = False,
):
  if set_logging:
    setup_logging(log_level)

  if not (server := retry_until_found(name, host, uuid, wait, retry_wait)):
    device = get_name(name, host, uuid)
    raise NoDevicesFound(device)

  server.adapter.set_icon(icon)
  server.loop(background=background)


def run_safe(args: Args):
  try:
    run_server(*args)

  except NoDevicesFound as e:
    log.error(f'Device {e} not found.')
    quit(Rc.NO_DEVICE)



from __future__ import annotations

from typing import Final


# module metadata
__author__: Final[str] = 'Alex DeLorenzo <alex@alexdelorenzo.dev>'
__license__: Final[str] = 'AGPL-3.0'
__copyright__: Final[str] = \
  f'Copyright 2023 {__author__}. Licensed under terms of the {__license__}.'
__version__: Final[str] = '0.14.0'

NAME: Final[str] = 'cast_control'
SHORT_NAME: Final[str] = 'castctl'
DESCRIPTION: Final[str] = '📺 Control Chromecasts from Linux and D-Bus'
HOMEPAGE: Final[str] = "https://github.com/alexdelorenzo/cast_control"
TITLE: Final[str] = 'Cast Control'

ENTRYPOINT_NAME: Final[str] = 'cli'
CLI_MODULE_NAME: Final[str] = f'{NAME}.app.cli'


# packaging metadata
CMD_PT: Final[str] = f'{CLI_MODULE_NAME}:{ENTRYPOINT_NAME}'
PY_VERSION: Final[str] = '>=3.12'
PKGS: Final[list[str]] = [
  NAME,
  f'{NAME}.app',
  f'{NAME}.device',
]

PROJECT_URLS: Final[dict[str, str]] = {
  'Homepage': 'https://alexdelorenzo.dev/',
  'Source': HOMEPAGE
}

ASSET_DIRS: Final[list[str]] = [
  'assets/*.desktop',
  'assets/icon/cc-*.svg',
  'assets/icon/*.yml',
]

CONSOLE_SCRIPTS: Final[list[str]] = [
  f'{NAME} = {CMD_PT}',
  f'{SHORT_NAME} = {CMD_PT}',
]

PKG_DATA: Final[dict[str, list[str]]] = {
  NAME: ASSET_DIRS
}

ENTRY_POINTS: Final[dict[str, list[str]]] = {
  'console_scripts': CONSOLE_SCRIPTS
}



from __future__ import annotations

from typing import Protocol, override, runtime_checkable

from mpris_server import DbusObj, LoopStatus, Metadata, Microseconds, Paths, PlayState, Rate, Track, \
  ValidMetadata, Volume
from pychromecast.controllers.media import MediaController, MediaStatus
from pychromecast.controllers.receiver import CastStatus
from pychromecast.socket_client import ConnectionStatus

from .base import DEFAULT_ICON, Device, NAME
from .device.base import CachedIcon, Controllers, Titles


@runtime_checkable
class CliIntegration(Protocol):
  def set_icon(self, lighter: bool = False): ...


@runtime_checkable
class ListenerIntegration(Protocol):
  def on_new_status(self, *args, **kwargs):
    """Callback for event listener"""


@runtime_checkable
class ModuleIntegration(Protocol):
  def get_duration(self) -> Microseconds: ...


@runtime_checkable
class Statuses(Protocol):
  @property
  def cast_status(self) -> CastStatus | None: ...

  @property
  def connection_status(self) -> ConnectionStatus | None: ...

  @property
  def media_controller(self) -> MediaController: ...

  @property
  def media_status(self) -> MediaStatus | None: ...


@runtime_checkable
class Properties(Protocol):
  device: Device
  controllers: Controllers

  cached_icon: CachedIcon | None = None
  light_icon: bool = DEFAULT_ICON

  @property
  def name(self) -> str:
    return self.device.name or NAME

  @property
  def is_youtube(self) -> bool: ...

  @property
  def titles(self) -> Titles: ...


@runtime_checkable
class RootAdapterIntegration(Protocol):
  def can_quit(self) -> bool: ...

  def get_desktop_entry(self) -> str: ...

  def get_mime_types(self) -> list[str]: ...

  def get_uri_schemes(self) -> list[str]: ...

  def has_tracklist(self) -> bool: ...

  def quit(self): ...


@runtime_checkable
class TrackListAdapterIntegration(Protocol):
  def add_track(self, uri: str, after_track: DbusObj, set_as_current: bool): ...

  def can_edit_tracks(self) -> bool: ...

  def get_tracks(self) -> list[DbusObj]: ...


@runtime_checkable
class PlayerAdapterIntegration(Protocol):
  def can_control(self) -> bool: ...

  def can_go_next(self) -> bool: ...

  def can_go_previous(self) -> bool: ...

  def can_pause(self) -> bool: ...

  def can_play(self) -> bool: ...

  def can_seek(self) -> bool: ...

  def get_art_url(self, track: int = None) -> str: ...

  def get_current_position(self) -> Microseconds: ...

  def get_current_track(self) -> Track: ...

  def get_next_track(self) -> Track: ...

  def get_playstate(self) -> PlayState: ...

  def get_previous_track(self) -> Track: ...

  def get_rate(self) -> Rate: ...

  def get_shuffle(self) -> bool: ...

  def get_stream_title(self) -> str: ...

  def get_volume(self) -> Volume: ...

  def is_mute(self) -> bool: ...

  def is_playlist(self) -> bool: ...

  def is_repeating(self) -> bool: ...

  def metadata(self) -> Metadata: ...

  def next(self): ...

  def open_uri(self, uri: str): ...

  def pause(self): ...

  def play(self): ...

  def previous(self): ...

  def resume(self): ...

  def seek(self, time: Microseconds, track_id: DbusObj | None = None): ...

  def set_icon(self, lighter: bool = False): ...

  def set_loop_status(self, value: LoopStatus): ...

  def set_mute(self, value: bool): ...

  def set_rate(self, value: Rate): ...

  def set_repeating(self, value: bool): ...

  def set_shuffle(self, value: bool): ...

  def set_volume(self, value: Volume): ...

  def stop(self): ...


@runtime_checkable
class AdapterIntegration(Protocol):
  def add_track(self, uri: str, after_track: DbusObj, set_as_current: bool): ...

  def can_control(self) -> bool: ...

  def can_edit_tracks(self) -> bool: ...

  def can_pause(self) -> bool: ...

  def can_play(self) -> bool: ...

  def can_play_next(self) -> bool: ...

  def can_play_prev(self) -> bool: ...

  def can_quit(self) -> bool: ...

  def can_seek(self) -> bool: ...

  def get_art_url(self, track: int | None = None) -> str: ...

  def get_desktop_entry(self) -> Paths: ...

  def get_playstate(self) -> PlayState: ...

  def get_rate(self) -> Rate: ...

  def get_shuffle(self) -> bool: ...

  def get_stream_title(self) -> str: ...

  def get_tracks(self) -> list[DbusObj]: ...

  def get_volume(self) -> Volume: ...

  def has_tracklist(self) -> bool: ...

  def has_current_time(self) -> bool: ...

  def is_mute(self) -> bool: ...

  def is_playlist(self) -> bool: ...

  def is_repeating(self) -> bool: ...

  def metadata(self) -> ValidMetadata: ...

  def next(self): ...

  def open_uri(self, uri: str): ...

  def pause(self): ...

  def play(self): ...

  def previous(self): ...

  def quit(self): ...

  def resume(self): ...

  def seek(self, time: Microseconds, track_id: DbusObj | None = None): ...

  def set_loop_status(self, value: LoopStatus): ...

  def set_mute(self, value: bool): ...

  def set_rate(self, value: Rate): ...

  def set_repeating(self, value: bool): ...

  def set_shuffle(self, value: bool): ...

  def set_volume(self, value: Volume): ...

  def stop(self): ...


@runtime_checkable
class Wrapper(
  AdapterIntegration,
  CliIntegration,
  ListenerIntegration,
  ModuleIntegration,
  Properties,
  Statuses,
  Protocol
):
  pass


@runtime_checkable
class DeviceIntegration[W: Wrapper](CliIntegration, ListenerIntegration, ModuleIntegration, Protocol):
  wrapper: W

  @override
  def get_duration(self) -> Microseconds:
    return self.wrapper.get_duration()

  @override
  def on_new_status(self, *args, **kwargs):
    self.wrapper.on_new_status(*args, **kwargs)

  @override
  def set_icon(self, lighter: bool = False):
    self.wrapper.set_icon(lighter)



from __future__ import annotations

from collections import deque
from collections.abc import Iterable, Iterator
from enum import StrEnum
from itertools import chain
from typing import Any, Final, NamedTuple, Self, TYPE_CHECKING
from urllib.parse import ParseResult, parse_qs, urlparse

from iteration_utilities import unique_everseen
from pychromecast.controllers.bbciplayer import BbcIplayerController
from pychromecast.controllers.bbcsounds import BbcSoundsController
from pychromecast.controllers.bubbleupnp import BubbleUPNPController
from pychromecast.controllers.dashcast import DashCastController
from pychromecast.controllers.homeassistant_media import HomeAssistantMediaController
from pychromecast.controllers.media import DefaultMediaReceiverController
from pychromecast.controllers.multizone import MultizoneController
from pychromecast.controllers.plex import PlexController
from pychromecast.controllers.receiver import ReceiverController
from pychromecast.controllers.supla import SuplaController
from pychromecast.controllers.yleareena import YleAreenaController
from pychromecast.controllers.youtube import YouTubeController
from validators import url

from ..base import Device, MediaType

if TYPE_CHECKING:
  from ..protocols import Wrapper


URL_PROTO: Final[str] = 'https'
SKIP_FIRST: Final[slice] = slice(1, None)


class CachedIcon(NamedTuple):
  url: str
  app_id: str | None = None
  title: str | None = None


class Controllers(NamedTuple):
  bbc_ip: BbcIplayerController | None = None
  bbc_sound: BbcSoundsController | None = None
  bubble: BubbleUPNPController | None = None
  dash: DashCastController | None = None
  default: DefaultMediaReceiverController | None = None
  ha_media: HomeAssistantMediaController | None = None
  multizone: MultizoneController | None = None
  plex: PlexController | None = None
  receiver: ReceiverController | None = None
  supla: SuplaController | None = None
  yle: YleAreenaController | None = None
  youtube: YouTubeController | None = None

  # plex_api: PlexApiController | None = None
  # ha: HomeAssistantController | None= None

  @classmethod
  def new(cls: type[Self], device: Device | None) -> Self:
    return cls(
      BbcIplayerController(),
      BbcSoundsController(),
      BubbleUPNPController(),
      DashCastController(),
      DefaultMediaReceiverController(),
      HomeAssistantMediaController(),
      MultizoneController(device.uuid) if device else None,
      PlexController(),
      ReceiverController(),
      SuplaController(),
      YleAreenaController(),
      YouTubeController(),
      # HomeAssistantController(),
    )

  def register(self, device: Device):
    for controller in self:
      if controller:
        device.register_handler(controller)


class Titles(NamedTuple):
  title: str | None = None
  artist: str | None = None
  album: str | None = None
  comments: str | None = None


class TitlesBuilder(Iterable[str]):
  title: str | None = None
  artist: str | None = None
  album: str | None = None
  comments: str | None = None

  _titles: deque[str]

  def __init__(
    self,
    *titles: str,
    title: str | None = None,
    artist: str | None = None,
    album: str | None = None,
    comments: str | None = None,
  ):
    self._titles = deque(titles)
    self.set(title=title, artist=artist, album=album, comments=comments)

  def __bool__(self) -> bool:
    return any(self)

  def __contains__(self, value: Any) -> bool:
    return value in iter(self)

  def __iter__(self) -> Iterator[str]:
    titles = chain(self.titles, self._titles)
    return filter(bool, titles)

  def __len__(self) -> int:
    return len(tuple(self))

  def __repr__(self) -> str:
    return repr(self.build())

  @property
  def titles(self) -> tuple[str | None, ...]:
    return self.title, self.artist, self.album, self.comments

  def add(self, *titles: str):
    titles = (title for title in titles if title and title not in self)
    self._titles.extend(titles)

  def set(
    self,
    *,
    title: str | None = None,
    artist: str | None = None,
    album: str | None = None,
    comments: str | None = None,
    overwrite: bool = True,
  ):
    if title:
      if overwrite or not self.title:
        self.title = title

      elif title not in self._titles:
        self.add(title)

    if artist:
      if overwrite or not self.artist:
        self.artist = artist

      else:
        self.add(artist)

    if album:
      if overwrite or not self.album:
        self.album = album

      else:
        self.add(album)

    if comments:
      if overwrite or not self.comments:
        self.comments = comments

      else:
        self.add(comments)

  def build(self) -> Titles:
    titles: list[str] = []
    rest: deque[str] = deque(unique_everseen(self._titles))

    for item in self.titles:
      titles.append(item if item else rest.popleft() if rest else None)

    return Titles(*titles)


class YoutubeUrl(StrEnum):
  long = 'youtube.com'
  short = 'youtu.be'

  watch_endpoint = 'watch'
  playlist_endpoint = 'playlist'

  video_query = 'v'
  playlist_query = 'list'

  video = f'{URL_PROTO}://{long}/{watch_endpoint}?{video_query}='
  playlist = f'{URL_PROTO}://{long}/{playlist_endpoint}?{playlist_query}='

  @classmethod
  def domain(cls: type[Self], uri: str | ParseResult) -> Self | None:
    match get_domain(uri):
      case cls.long:
        return cls.long

      case cls.short:
        return cls.short

    return None

  @classmethod
  def get_content_id(cls: type[Self], uri: str | None) -> str | None:
    return get_content_id(uri)

  @classmethod
  def get_url(cls: type[Self], video_id: str | None = None, playlist_id: str | None = None) -> str | None:
    if video_id:
      return f"{cls.video}{video_id}"

    elif playlist_id:
      return f"{cls.playlist}{playlist_id}"

    return None

  @classmethod
  def is_youtube(cls: type[Self], uri: str | None) -> bool:
    if not uri:
      return False

    uri = uri.casefold()

    return get_domain(uri) in cls

  @classmethod
  def type(cls: type[Self], uri: str | None) -> Self | None:
    if not (which := cls.which(uri)):
      return None

    if cls.watch_endpoint in uri:
      return cls.video

    elif cls.playlist_endpoint in uri:
      return cls.playlist

    return which

  @classmethod
  def which(cls: type[Self], uri: str | None) -> Self | None:
    if not cls.is_youtube(uri):
      return None

    if cls.long in uri:
      return cls.long

    elif cls.short in uri:
      return cls.short

    return None


def get_domain(uri: str | ParseResult) -> str | None:
  if not url(uri):
    return None

  if isinstance(uri, str):
    uri = urlparse(uri)

  *_, name, tld = uri.netloc.split(".")

  return f"{name}.{tld}"


def get_content_id(uri: str) -> str | None:
  if not url(uri) or not YoutubeUrl.is_youtube(uri):
    return None

  parsed = urlparse(uri)
  content_id: str | None = None

  match YoutubeUrl.domain(uri), YoutubeUrl.type(uri):
    case YoutubeUrl.long, YoutubeUrl.video:
      qs = parse_qs(parsed.query)
      [content_id] = qs[YoutubeUrl.video_query]

    case YoutubeUrl.long, YoutubeUrl.playlist:
      qs = parse_qs(parsed.query)
      [content_id] = qs[YoutubeUrl.playlist_query]

    case YoutubeUrl.short, YoutubeUrl.video | YoutubeUrl.playlist:
      content_id = parsed.path[SKIP_FIRST]

  return content_id


def get_media_type(wrapper: Wrapper) -> MediaType | None:
  if not (status := wrapper.media_status):
    return None

  if status.media_is_movie:
    return MediaType.MOVIE

  elif status.media_is_tvshow:
    return MediaType.TVSHOW

  elif status.media_is_photo:
    return MediaType.PHOTO

  elif status.media_is_musictrack:
    return MediaType.MUSICTRACK

  elif status.media_is_generic:
    return MediaType.GENERIC

  return None



from __future__ import annotations

import logging
from asyncio import gather, run
from collections.abc import Callable
from functools import cache, wraps
from os import stat_result
from pathlib import Path

from aiopath import AsyncPath
from rich.logging import RichHandler

from ..base import DARK_END, DARK_ICON, DATA_DIR, DESKTOP_NAME, DESKTOP_SUFFIX, DESKTOP_TEMPLATE, LIGHT_END, \
  LIGHT_ICON, LOG_FILE_MODE, LOG_LEVEL, NAME, PATHS, SRC_DIR, USER_DIRS, singleton


type Decoratable[**P, T] = Callable[P, T]
type Decorated[**P, T] = Callable[P, T]


def setup_logging(
  level: str = LOG_LEVEL,
  file: Path | None = None,
):
  level = level.upper()

  if file:
    create_user_dirs()

    logging.basicConfig(
      level=level,
      filename=file,
      filemode=LOG_FILE_MODE,
    )

  else:
    handlers = [RichHandler(rich_tracebacks=True)]
    logging.basicConfig(level=level, handlers=handlers)


# check for user dirs and create them asynchronously
async def _create_user_dirs():
  await PATHS.create_user_paths()

  paths = map(AsyncPath, USER_DIRS)
  coros = (path.mkdir(parents=True, exist_ok=True) for path in paths)

  await gather(*coros)


@singleton
def create_user_dirs():
  run(_create_user_dirs())


def ensure_user_dirs_exist[**P, T](func: Decoratable) -> Decorated:
  @wraps(func)
  def new_func(*args: P.args, **kwargs: P.kwargs) -> T:
    create_user_dirs()
    return func(*args, **kwargs)

  return new_func


def get_stat(file: Path) -> stat_result:
  return file.stat()


@singleton
def get_src_stat() -> stat_result:
  return get_stat(SRC_DIR)


@singleton
def get_template() -> list[str]:
  return DESKTOP_TEMPLATE \
    .read_text() \
    .splitlines()


def is_older_than_module(other: Path) -> bool:
  src_stat = get_src_stat()
  other_stat = get_stat(other)

  return src_stat.st_ctime > other_stat.st_ctime


def get_paths(light_icon: bool = True) -> tuple[Path, Path]:
  icon_path = LIGHT_ICON if light_icon else DARK_ICON
  name_suffix = LIGHT_END if light_icon else DARK_END
  new_name = f'{NAME}{name_suffix}{DESKTOP_SUFFIX}'
  desktop_path = DATA_DIR / new_name

  return desktop_path, icon_path


@cache
def new_file_from_template(file: Path, icon_path: Path):
  *lines, name, icon = get_template()
  name += DESKTOP_NAME
  icon += str(icon_path)
  lines = (*lines, name, icon)
  text = '\n'.join(lines)

  file.write_text(text)


@cache
@ensure_user_dirs_exist
def create_desktop_file(light_icon: bool = True) -> Path:
  file, icon = get_paths(light_icon)

  if not file.exists() or is_older_than_module(file):
    new_file_from_template(file, icon)

  return file



from __future__ import annotations

from collections.abc import Callable
from decimal import Context, Decimal, ROUND_HALF_UP, getcontext
from enum import IntEnum, StrEnum, auto
from functools import lru_cache
from pathlib import Path
from typing import Final

from app_paths import AsyncAppPaths, get_paths
from pychromecast import Chromecast
from pychromecast.controllers.media import MediaStatus
from pychromecast.controllers.receiver import CastStatus, LaunchFailure
from pychromecast.socket_client import ConnectionStatus

from . import NAME, __author__, __version__


Seconds = Decimal

DESKTOP_NAME: Final[str] = 'Cast Control'
LOG_LEVEL: Final[str] = 'WARN'

NO_DURATION: Final[int] = 0
NO_DELTA: Final[int] = 0
NO_DEVICE_NAME: Final[str] = 'NO_NAME'
NO_STR: Final[str] = ''
NO_PORT: Final[int | None] = None

# older Python requires an explicit
# maxsize param for lru_cache()
SINGLETON: Final[int] = 1

YOUTUBE: Final[str] = 'YouTube'

US_IN_SEC: Final[int] = 1_000_000  # seconds to microseconds
DEFAULT_TRACK: Final[str] = '/track/1'
DEFAULT_DISC_NO: Final[int] = 1

DEFAULT_RETRY_WAIT: Final[Seconds] = Seconds(5.0)
DEFAULT_WAIT: Final[Seconds] = Seconds(30)
DEFAULT_DEVICE_NAME: Final[str] = DESKTOP_NAME
DEFAULT_NO_DEVICE_NAME: Final[str] = 'Device'

LOG_FILE_MODE: Final[str] = 'w'  # create a new log on service start
DEFAULT_ICON: Final[bool] = False
DEFAULT_SET_LOG: Final[bool] = False

DESKTOP_SUFFIX: Final[str] = '.desktop'
NO_DESKTOP_FILE: Final[str] = ''

ARGS_STEM: Final[str] = '-args'
LIGHT_END: Final[str] = '-light'
DARK_END: Final[str] = '-dark'

PATHS: Final[AsyncAppPaths] = get_paths(
  NAME,
  __author__,
  __version__,
  is_async=True
)
DATA_DIR: Final[Path] = Path(PATHS.user_data_path)
LOG_DIR: Final[Path] = Path(PATHS.user_log_path)
STATE_DIR: Final[Path] = Path(PATHS.user_state_path)

USER_DIRS: Final[tuple[Path, ...]] = DATA_DIR, LOG_DIR, STATE_DIR

PID: Final[Path] = STATE_DIR / f'{NAME}.pid'
ARGS: Final[Path] = STATE_DIR / f'service{ARGS_STEM}.tmp'
LOG: Final[Path] = LOG_DIR / f'{NAME}.log'

SRC_DIR: Final[Path] = Path(__file__).parent
ASSETS_DIR: Final[Path] = SRC_DIR / 'assets'
DESKTOP_TEMPLATE: Final[Path] = ASSETS_DIR / f'template{DESKTOP_SUFFIX}'

ICON_DIR: Final[Path] = ASSETS_DIR / 'icon'
DARK_SVG: Final[Path] = ICON_DIR / 'cc-black.svg'
LIGHT_SVG: Final[Path] = ICON_DIR / 'cc-white.svg'
TEMPLATE_SVG: Final[Path] = ICON_DIR / 'cc-template.svg'

LIGHT_ICON = LIGHT_THUMB = LIGHT_SVG
DEFAULT_THUMB = DARK_ICON = DARK_SVG

PRECISION: Final[int] = 4
CONTEXT: Final[Context] = getcontext()
CONTEXT.prec = PRECISION
CONTEXT.rounding = ROUND_HALF_UP

Device = Chromecast
Status = MediaStatus | CastStatus | ConnectionStatus | LaunchFailure

type Decorated[** P, T] = Callable[P, T]
type Decoratable[** P, T] = Callable[P, T]
type Decorator[** P, T] = Callable[[Decoratable], Decorated]


class NoDevicesFound(Exception):
  pass


class MediaType(StrEnum):
  GENERIC = auto()
  MOVIE = auto()
  MUSICTRACK = auto()
  PHOTO = auto()
  TVSHOW = auto()


class Rc(IntEnum):
  OK = 0
  NO_DEVICE = auto()
  NOT_RUNNING = auto()


singleton: Final[Decorator] = lru_cache(SINGLETON)



from __future__ import annotations

import pickle
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import NamedTuple
from uuid import UUID

from daemons.prefab.run import RunDaemon

from .state import setup_logging
from ..base import ARGS, ARGS_STEM, DEFAULT_ICON, DEFAULT_NO_DEVICE_NAME, DEFAULT_RETRY_WAIT, DEFAULT_SET_LOG, \
  DEFAULT_WAIT, LOG, LOG_LEVEL, PID, Seconds


class MprisDaemon[**P, T](RunDaemon):
  target: Callable[P, T] | None = None
  args: Args | None = None
  _logging: str | None = None

  @property
  def logging(self) -> str | None:
    return self._logging

  @logging.setter
  def logging(self, val: str | None):
    self._logging = val

  def set_target[**P, T](
    self,
    func: Callable[P, T] | None = None,
    *args,
    **kwargs
  ):
    if not func:
      self.target = None
      return

    self.target = partial(func, *args, **kwargs)

  def set_target_via_args[**P, T](
    self,
    func: Callable[P, T] | None = None,
    args: Args | None = None
  ):
    if not func:
      self.target = None
      return

    self.args = args
    self.logging = args.set_logging
    self.target = partial(func, args)

  def setup_logging(self):
    if self.args:
      level = self.args.log_level

    else:
      level = self.logging

    setup_logging(level, file=LOG)

  def run(self):
    if not self.target:
      return

    self.setup_logging()
    self.target()


class Args(NamedTuple):
  name: str | None = None
  host: str | None = None
  uuid: UUID | str | None = None
  wait: Seconds | None = DEFAULT_WAIT
  retry_wait: Seconds | None = DEFAULT_RETRY_WAIT
  icon: bool = DEFAULT_ICON
  log_level: str = LOG_LEVEL
  set_logging: bool = DEFAULT_SET_LOG
  background: bool = False

  @staticmethod
  def load(identifier: str | None = None) -> Args | None:
    if identifier:
      args = ARGS.with_stem(f'{identifier}{ARGS_STEM}')

    else:
      args = ARGS

    if args.exists():
      dump = args.read_bytes()
      return pickle.loads(dump)

    return None

  @staticmethod
  def delete():
    if ARGS.exists():
      ARGS.unlink()

  def save(self) -> Path:
    dump = pickle.dumps(self)
    ARGS.write_bytes(dump)

    return ARGS

  @property
  def file(self) -> Path:
    name, host, uuid, *_ = self
    device = get_name(name, host, uuid)

    return ARGS.with_stem(f'{device}{ARGS_STEM}')


def get_daemon[**P, T](
  func: Callable[P, T] | None = None,
  *args,
  _pidfile: str = str(PID),
  **kwargs,
) -> MprisDaemon:
  daemon = MprisDaemon(pidfile=_pidfile)
  daemon.set_target(func, *args, **kwargs)

  return daemon


def get_daemon_from_args[**P, T](
  func: Callable[P, T] | None = None,
  args: Args | None = None,
  _pidfile: str = str(PID),
) -> MprisDaemon:
  daemon = MprisDaemon(pidfile=_pidfile)
  daemon.set_target_via_args(func, args)

  return daemon


def get_name(name: str | None, host: str | None, uuid: UUID | str | None) -> str:
  return name or host or uuid or DEFAULT_NO_DEVICE_NAME



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Final, Self, override

from mpris_server import EventAdapter, Server
from pychromecast.controllers.media import MediaStatus, MediaStatusListener
from pychromecast.controllers.receiver import CastStatus, CastStatusListener, LaunchErrorListener, LaunchFailure
from pychromecast.socket_client import ConnectionStatus, ConnectionStatusListener

from ..adapter import DeviceAdapter
from ..base import Device, Status


log: Final[logging.Logger] = logging.getLogger(__name__)


# status with volume attributes
VolumeStatus = MediaStatus | CastStatus


class BaseEventListener(
  CastStatusListener,
  ConnectionStatusListener,
  LaunchErrorListener,
  MediaStatusListener,
  ABC
):
  """Event listeners that conform to PyChromecast's API"""

  @override
  @abstractmethod
  def load_media_failed(self, item: int, error_code: int):
    pass

  @override
  @abstractmethod
  def new_cast_status(self, status: CastStatus):
    pass

  @override
  @abstractmethod
  def new_connection_status(self, status: ConnectionStatus):
    pass

  @override
  @abstractmethod
  def new_launch_error(self, status: LaunchFailure):
    pass

  @override
  @abstractmethod
  def new_media_status(self, status: MediaStatus):
    pass


class BaseEventAdapter(EventAdapter):
  server: Server
  device: Device

  name: str
  adapter: DeviceAdapter | None

  @override
  def __init__(self, server: Server, device: Device):
    self.server = server
    self.device = device

    self.name = self.device.name
    self.adapter = self.server.adapter

    super().__init__(
      root=self.server.root,
      player=self.server.player,
      tracklist=self.server.tracklist,
      playlists=self.server.playlists,
    )

  @classmethod
  def register(cls: type[Self], server: Server, device: Device) -> Self:
    events = cls(server, device)
    events.set_and_register()

    return events

  def set_and_register(self):
    self.server.set_event_adapter(self)


class EventListener(BaseEventAdapter, BaseEventListener):
  def _update_volume(self, status: Status | None = None):
    if isinstance(status, VolumeStatus):
      self.on_volume()

  def _update_metadata(self, status: Status | None = None):
    self._update_volume(status)

    # wire up local integration with mpris
    self.adapter.on_new_status()

    # wire up mpris_server with cc events
    self.on_root_all()
    self.on_player_all()
    self.on_tracklist_all()
    # self.on_playlists_all()
    # self.emit_all()

  @override
  def set_and_register(self):
    super().set_and_register()
    register_event_listener(self, self.device)

  @override
  def load_media_failed(self, item: int, error_code: int):
    log.error(f'Load media failed: {error_code=}, {item=}')
    self._update_metadata()

  @override
  def new_cast_status(self, status: CastStatus):
    log.debug(f'Handling new cast status: {status}')
    self._update_metadata(status)

  @override
  def new_connection_status(self, status: ConnectionStatus):
    log.info(f'Handling new connection status: {status}')
    self._update_metadata(status)

  @override
  def new_launch_error(self, status: LaunchFailure):
    log.error(f'Handling new launch error: {status}')
    self._update_metadata(status)

  @override
  def new_media_status(self, status: MediaStatus):
    log.debug(f'Handling new media status: {status}')
    self._update_metadata(status)


def register_event_listener[E: BaseEventListener](events: E, device: Device):
  device.register_connection_listener(events)
  device.register_launch_error_listener(events)
  device.register_status_listener(events)
  device.media_controller.register_status_listener(events)



from __future__ import annotations

import logging
from typing import Final, NamedTuple

import click

from .daemon import Args, MprisDaemon, get_daemon, get_daemon_from_args
from .run import run_safe
from .. import CLI_MODULE_NAME, ENTRYPOINT_NAME, HOMEPAGE, __copyright__, __version__
from ..base import DEFAULT_DEVICE_NAME, DEFAULT_RETRY_WAIT, LOG, LOG_LEVEL, NAME, Rc, Seconds


assert __name__ == CLI_MODULE_NAME


log: Final[logging.Logger] = logging.getLogger(__name__)

LOG_MODE: Final[str] = 'r'
LOG_END: Final[str] = ''

VERSION_INFO: Final[str] = f'{NAME} v{__version__}'

NOT_RUNNING_MSG: Final[str] = "Daemon isn't running."
HELP: Final[str] = f'''
Control casting devices via Linux media controls and desktops.

This daemon connects your casting device to the D-Bus media player interface (MPRIS).

See {HOMEPAGE} for more information.
'''


type KwargsVal = bool | str | int | float | click.ParamType


class CliArgs(NamedTuple):
  args: tuple[str, ...]
  kwargs: dict[str, KwargsVal]


NAME_ARGS: Final[CliArgs] = CliArgs(
  args=('--name', '-n'),
  kwargs=dict(
    default=DEFAULT_DEVICE_NAME,
    show_default=True,
    type=click.STRING,
    help="Connect to a device via its name, otherwise control the first device found."
  )
)

HOST_ARGS: Final[CliArgs] = CliArgs(
  args=('--host', '-h'),
  kwargs=dict(
    default=None,
    show_default=True,
    type=click.STRING,
    help="Connect to a device via its hostname or IP address, otherwise control the first device found."
  )
)

UUID_ARGS: Final[CliArgs] = CliArgs(
  args=('--uuid', '-u'),
  kwargs=dict(
    default=None,
    show_default=True,
    type=click.STRING,
    help="Connect to a device via its UUID, otherwise control the first device found."
  )
)

WAIT_ARGS: Final[CliArgs] = CliArgs(
  args=('--wait', '-w'),
  kwargs=dict(
    default=None,
    show_default=True,
    type=click.FLOAT,
    help="Seconds to wait between trying to make initial successful connections to a device."
  )
)

RETRY_ARGS: Final[CliArgs] = CliArgs(
  args=('--retry-wait', '-r'),
  kwargs=dict(
    default=DEFAULT_RETRY_WAIT,
    show_default=True,
    type=click.FLOAT,
    help="Seconds to wait between reconnection attempts if a successful connection is interrupted."
  )
)

ICON_ARGS: Final[CliArgs] = CliArgs(
  args=('--icon', '-i'),
  kwargs=dict(
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
    help="Use a lighter icon instead of the dark icon. The lighter icon goes well with dark themes."
  )
)

LOG_ARGS: Final[CliArgs] = CliArgs(
  args=('--log-level', '-l'),
  kwargs=dict(
    default=LOG_LEVEL,
    show_default=True,
    type=click.STRING,
    help='Set the debugging log level.'
  )
)


# see https://alexdelorenzo.dev/notes/click
class OrderAsCreated(click.Group):
  """List `click` commands in the order they're declared."""

  def list_commands(self, ctx: click.Context) -> list[str]:
    return list(self.commands)


@click.group(
  help=HELP,
  invoke_without_command=True
)
@click.option(
  '--license', '-L',
  is_flag=True, default=False, type=click.BOOL,
  help="Show license and copyright information."
)
@click.option(
  '--version', '-V',
  is_flag=True, default=False, type=click.BOOL,
  help="Show version information."
)
@click.pass_context
def cli(
  ctx: click.Context,
  license: bool,
  version: bool,
):
  if license:
    click.echo(__copyright__)

  if version:
    click.echo(VERSION_INFO)

  if license or version:
    quit(Rc.OK)

  elif ctx.invoked_subcommand:
    return

  help: str = cli.get_help(ctx)
  click.echo(help)


assert cli.name == ENTRYPOINT_NAME


@cli.command(help='Connect to a device and run the service in the foreground.')
@click.option(*NAME_ARGS.args, **NAME_ARGS.kwargs)
@click.option(*HOST_ARGS.args, **HOST_ARGS.kwargs)
@click.option(*UUID_ARGS.args, **UUID_ARGS.kwargs)
@click.option(*WAIT_ARGS.args, **WAIT_ARGS.kwargs)
@click.option(*RETRY_ARGS.args, **RETRY_ARGS.kwargs)
@click.option(*ICON_ARGS.args, **ICON_ARGS.kwargs)
@click.option(*LOG_ARGS.args, **LOG_ARGS.kwargs)
def connect(
  name: str | None,
  host: str | None,
  uuid: str | None,
  wait: Seconds | None,
  retry_wait: Seconds | None,
  icon: bool,
  log_level: str
):
  args = Args(name, host, uuid, wait, retry_wait, icon, log_level, set_logging=True)
  run_safe(args)


@cli.group(
  cls=OrderAsCreated,
  help='Connect, disconnect or reconnect the background service to or from your device.',
)
def service():
  pass


@service.command(help='Connect the background service to a device.')
@click.option(*NAME_ARGS.args, **NAME_ARGS.kwargs)
@click.option(*HOST_ARGS.args, **HOST_ARGS.kwargs)
@click.option(*UUID_ARGS.args, **UUID_ARGS.kwargs)
@click.option(*WAIT_ARGS.args, **WAIT_ARGS.kwargs)
@click.option(*RETRY_ARGS.args, **RETRY_ARGS.kwargs)
@click.option(*ICON_ARGS.args, **ICON_ARGS.kwargs)
@click.option(*LOG_ARGS.args, **LOG_ARGS.kwargs)
def connect(
  name: str | None,
  host: str | None,
  uuid: str | None,
  wait: Seconds | None,
  retry_wait: Seconds | None,
  icon: bool,
  log_level: str
):
  args = Args(name, host, uuid, wait, retry_wait, icon, log_level)
  args.save()

  try:
    daemon = get_daemon_from_args(run_safe, args)
    daemon.start()

  except Exception as e:
    log.exception(e)
    log.error("Error launching daemon.")

    args.delete()


@service.command(help='Disconnect the background service from the device.')
def disconnect():
  daemon = get_daemon()

  if not daemon.pid:
    log.error(NOT_RUNNING_MSG)
    quit(Rc.NOT_RUNNING)

  daemon.stop()
  Args.delete()


@service.command(help='Reconnect the background service to the device.')
def reconnect():
  daemon: MprisDaemon | None = None
  args = Args.load()

  if args:
    daemon = get_daemon_from_args(run_safe, args)

  if not args or not daemon.pid:
    log.error(NOT_RUNNING_MSG)
    quit(Rc.NOT_RUNNING)

  daemon.restart()


@service.command(help='Show the service log.')
def log():
  click.echo(f"<Log file: {LOG}>")

  # a large log could hang Python or the system
  # iterate over the file instead of using Path.read_text()
  with LOG.open(LOG_MODE) as file:
    for line in file:
      print(line, end=LOG_END)


if __name__ == "__main__":
  cli()



from __future__ import annotations

import logging
from decimal import Decimal
from mimetypes import guess_type
from typing import Final, override

from mpris_server import (
  Album, Artist, BEGINNING, DEFAULT_RATE, DbusObj, LoopStatus, MetadataObj, Microseconds, Paths, PlayState, Rate, Track,
  ValidMetadata, Volume, get_track_id,
)
from pychromecast.controllers.media import MediaController, MediaImage, MediaStatus
from pychromecast.controllers.receiver import CastStatus
from pychromecast.socket_client import ConnectionStatus
from validators import url

from .base import CachedIcon, Controllers, Titles, TitlesBuilder, YoutubeUrl
from .. import TITLE
from ..app.state import create_desktop_file, ensure_user_dirs_exist
from ..base import DEFAULT_DISC_NO, DEFAULT_THUMB, Device, \
  LIGHT_THUMB, NO_DELTA, NO_DESKTOP_FILE, \
  NO_DURATION, Seconds, US_IN_SEC, singleton
from ..protocols import CliIntegration, ListenerIntegration, ModuleIntegration, Wrapper


log: Final[logging.Logger] = logging.getLogger(__name__)


RESOLUTION: Final[int] = 1
MAX_TITLES: Final[int] = 3

NO_ARTIST: Final[str] = ''
NO_SUFFIX: Final[str] = ''

PREFIX_NOT_YOUTUBE: Final[str] = 'http'


class StatusMixin(Wrapper):
  @override
  @property
  def cast_status(self) -> CastStatus | None:
    return self.device.status or None

  @override
  @property
  def media_status(self) -> MediaStatus | None:
    return self.media_controller.status or None

  @override
  @property
  def connection_status(self) -> ConnectionStatus | None:
    return self.device.socket_client.receiver_controller.status or None

  @override
  @property
  def media_controller(self) -> MediaController:
    return self.device.media_controller


class ControllersMixin(Wrapper):
  controllers: Controllers

  @override
  def __init__(self):
    self._setup_controllers()
    super().__init__()

  def _setup_controllers(self):
    self.controllers = Controllers.new(self.device)
    self.controllers.register(self.device)

  def _launch_youtube(self):
    if not (youtube := self.controllers.youtube):
      return

    youtube.launch()

  def _play_youtube(self, video_id: str):
    if not (youtube := self.controllers.youtube):
      return

    if not youtube.is_active:
      self._launch_youtube()

    youtube.quick_play(video_id)

  @property
  def is_youtube(self) -> bool:
    if youtube := self.controllers.youtube:
      return youtube.is_active

    return False

  @override
  def open_uri(self, uri: str):
    if content_id := YoutubeUrl.get_content_id(uri):
      self._play_youtube(content_id)
      return

    mimetype, _ = guess_type(uri)
    self.media_controller.play_media(uri, mimetype)

  @override
  def add_track(self, uri: str, after_track: DbusObj, set_as_current: bool):
    if not (youtube := self.controllers.youtube):
      self.open_uri(uri)
      return

    if content_id := YoutubeUrl.get_content_id(uri):
      youtube.add_to_queue(content_id)

    if content_id and set_as_current:
      youtube.play_video(content_id)

    elif set_as_current:
      self.open_uri(uri)


class TitlesMixin(Wrapper):
  @override
  @property
  def titles(self) -> Titles:
    titles: TitlesBuilder = TitlesBuilder()

    if title := self.media_controller.title:
      titles.set(title=title)

    if (subtitle := self.get_subtitle()) and self.is_youtube:
      titles.set(artist=subtitle)

    elif subtitle:
      titles.add(subtitle)

    if status := self.media_status:
      if title := status.series_title:
        titles.set(title=title, overwrite=False)

      if artist := status.artist:
        titles.set(artist=artist)

      if album := status.album_name:
        titles.set(album=album)

    if app_name := self.device.app_display_name:
      if not titles.artist:
        titles.set(artist=app_name)

      elif not titles.album:
        titles.set(album=app_name)

      elif not titles.title:
        titles.set(title=app_name)

    titles.add(TITLE)

    return titles.build()

  def get_subtitle(self) -> str | None:
    if not (status := self.media_status) or not (metadata := status.media_metadata):
      return None

    if subtitle := metadata.get('subtitle'):
      return subtitle

    return None


class TimeMixin(Wrapper, ListenerIntegration, ModuleIntegration):
  _longest_duration: Microseconds | None

  @override
  def __init__(self):
    self._longest_duration = NO_DURATION
    super().__init__()

  def _reset_longest_duration(self):
    if not self.has_current_time():
      self._longest_duration = None

  @override
  def on_new_status(self, *args, **kwargs):
    self._reset_longest_duration()
    super().on_new_status(*args, **kwargs)

  @override
  @property
  def current_time(self) -> Seconds | None:
    if not (status := self.media_status):
      return None

    if time := status.adjusted_current_time or status.current_time:
      return Seconds(time)

    return None

  @override
  def get_duration(self) -> Microseconds:
    if (status := self.media_status) and (duration := status.duration) is not None:
      duration = Seconds(duration)
      duration_us = duration * US_IN_SEC

      return round(duration_us)

    current: Microseconds = self.get_current_position()
    longest: Microseconds = self._longest_duration

    if longest and longest > current:
      return longest

    elif current:
      self._longest_duration = current
      return current

    return NO_DURATION

  @override
  def get_current_position(self) -> Microseconds:
    position: Seconds | None = self.current_time

    if not position:
      return BEGINNING

    position_us = position * US_IN_SEC
    return round(position_us)

  @override
  def has_current_time(self) -> bool:
    current_time: Seconds | None = self.current_time

    if current_time is None:
      return False

    current_time = round(current_time, RESOLUTION)

    return current_time > BEGINNING

  @override
  def seek(self, time: Microseconds, *_):
    microseconds = Decimal(time)
    seconds: int = round(microseconds / US_IN_SEC)

    self.media_controller.seek(seconds)

  @override
  def get_rate(self) -> Rate:
    if not (status := self.media_status):
      return DEFAULT_RATE

    if rate := status.playback_rate:
      return rate

    return DEFAULT_RATE

  @override
  def set_rate(self, value: Rate):
    pass


class IconsMixin(Wrapper, CliIntegration):
  cached_icon: CachedIcon | None
  light_icon: bool

  def _set_cached_icon(self, url: str | None = None):
    if not url:
      self.cached_icon = None
      return

    app_id = self.device.app_id
    title, *_ = self.titles
    self.cached_icon = CachedIcon(url, app_id, title)

  def _can_use_cache(self) -> bool:
    if not (icon := self.cached_icon) or not icon.url:
      return False

    app_id = self.device.app_id
    title, *_ = self.titles

    return icon.app_id == app_id and icon.title == title

  def _get_icon_from_device(self) -> str | None:
    url: str | None

    if (status := self.media_status) and (images := status.images):
      first: MediaImage

      first, *_ = images
      url, *_ = first
      self._set_cached_icon(url)

      return url

    if (status := self.cast_status) and (url := status.icon_url):
      self._set_cached_icon(url)
      return url

    if not self._can_use_cache():
      return None

    if icon := self.cached_icon:
      return icon.url

    return None

  @ensure_user_dirs_exist
  def _get_default_icon(self) -> str:
    if self.light_icon:
      return str(LIGHT_THUMB)

    return str(DEFAULT_THUMB)

  @override
  def get_art_url(self, track: int | None = None) -> str:
    if icon := self._get_icon_from_device():
      return icon

    return self._get_default_icon()

  @override
  @singleton
  def get_desktop_entry(self) -> Paths:
    try:
      return create_desktop_file(self.light_icon)

    except Exception as e:
      log.exception(e)
      log.error("Couldn't load desktop file.")

      return NO_DESKTOP_FILE

  @override
  def set_icon(self, lighter: bool = False):
    self.light_icon: bool = lighter


class MetadataMixin(Wrapper):
  def _get_url(self) -> str | None:
    content_id: str | None = None

    if status := self.media_status:
      content_id = status.content_id

    if self._is_youtube_video(content_id):
      return YoutubeUrl.get_url(content_id)

    return content_id

  def _is_youtube_video(self, content_id: str | None) -> bool:
    if not (youtube := self.controllers.youtube):
      return False

    if not content_id or not youtube.is_active:
      return False

    return not content_id.startswith(PREFIX_NOT_YOUTUBE)

  @override
  def metadata(self) -> ValidMetadata:
    title, artist, album, comments = self.titles

    dbus_name: DbusObj = get_track_id(title)
    artists: list[str] = [artist] if artist else []
    comments: list[str] = [comments] if comments else []
    track_no: int | None = None

    if status := self.media_status:
      track_no = status.track

    return MetadataObj(
      album=album,
      album_artists=artists,
      art_url=self.get_art_url(),
      artists=artists,
      comments=comments,
      disc_number=DEFAULT_DISC_NO,
      length=self.get_duration(),
      title=title,
      track_id=dbus_name,
      track_number=track_no,
      url=self._get_url(),
    )

  @override
  def get_stream_title(self) -> str:
    if status := self.media_status:
      return status.title

    return self.titles.title

  @override
  def get_current_track(self) -> Track:
    title, artist, album, comments = self.titles

    dbus_name: DbusObj = get_track_id(title)
    artists: list[Artist] = [Artist(artist)] if artist else []
    track_no: int | None = None
    art_url = self.get_art_url()

    if status := self.media_status:
      track_no = status.track

    return Track(
      album=Album(art_url, artists, album),
      art_url=art_url,
      artists=artists,
      comments=[comments] if comments else [],
      disc_number=DEFAULT_DISC_NO,
      length=self.get_duration(),
      name=title,
      track_id=dbus_name,
      track_number=track_no,
    )


class PlaybackMixin(Wrapper):
  @override
  def get_playstate(self) -> PlayState:
    if self.media_controller.is_playing:
      return PlayState.PLAYING

    elif self.media_controller.is_paused:
      return PlayState.PAUSED

    return PlayState.STOPPED

  @override
  def is_repeating(self) -> bool:
    return False

  @override
  def is_playlist(self) -> bool:
    return self.can_play_next() or self.can_play_prev()

  @override
  def get_shuffle(self) -> bool:
    return False

  @override
  def set_shuffle(self, value: bool):
    pass

  @override
  def quit(self):
    self.device.quit_app()

  @override
  def next(self):
    self.media_controller.queue_next()

  @override
  def previous(self):
    self.media_controller.queue_prev()

  @override
  def pause(self):
    self.media_controller.pause()

  @override
  def resume(self):
    self.play()

  @override
  def stop(self):
    self.media_controller.stop()

  @override
  def play(self):
    self.media_controller.play()

  @override
  def set_repeating(self, value: bool):
    pass

  @override
  def set_loop_status(self, value: LoopStatus):
    pass


class VolumeMixin(Wrapper):
  @override
  def get_volume(self) -> Volume | None:
    if status := self.cast_status:
      return Volume(status.volume_level)

    return None

  @override
  def set_volume(self, value: Volume):
    if (current := self.get_volume()) is None:
      return

    volume = Volume(value)
    delta: float = float(volume - current)

    # can't adjust vol by 0
    if delta > NO_DELTA:  # vol up
      self.device.volume_up(delta)

    elif delta < NO_DELTA:
      self.device.volume_down(abs(delta))

  @override
  def is_mute(self) -> bool | None:
    if status := self.cast_status or self.media_status:
      return status.volume_muted

    return False

  @override
  def set_mute(self, value: bool):
    self.device.set_volume_muted(value)


class AbilitiesMixin(Wrapper):
  @override
  def can_quit(self) -> bool:
    return True

  @override
  def can_play(self) -> bool:
    state = self.get_playstate()

    return state is not PlayState.STOPPED

  @override
  def can_control(self) -> bool:
    return True
    # return self.can_play() or self.can_pause() \
    #   or self.can_play_next() or self.can_play_prev() \
    #   or self.can_seek()

  @override
  def can_edit_tracks(self) -> bool:
    return False

  @override
  def can_play_next(self) -> bool:
    if status := self.media_status:
      return status.supports_queue_next

    return False

  @override
  def can_play_prev(self) -> bool:
    if status := self.media_status:
      return status.supports_queue_prev

    return False

  @override
  def can_pause(self) -> bool:
    if status := self.media_status:
      return status.supports_pause

    return False

  @override
  def can_seek(self) -> bool:
    if status := self.media_status:
      return status.supports_seek

    return False


class TracklistMixin(Wrapper):
  @override
  def has_tracklist(self) -> bool:
    return bool(self.get_tracks())

  @override
  def get_tracks(self) -> list[DbusObj]:
    title, *_ = self.titles

    if title:
      return [get_track_id(title)]

    return []


class DeviceWrapper(
  AbilitiesMixin,
  ControllersMixin,
  IconsMixin,
  MetadataMixin,
  PlaybackMixin,
  StatusMixin,
  TimeMixin,
  TitlesMixin,
  TracklistMixin,
  VolumeMixin,
):
  """Wraps implementation details for device API"""

  @override
  def __init__(self, device: Device):
    self.device = device
    super().__init__()

  @override
  def __repr__(self) -> str:
    cls = type(self)

    return f'<{cls.__name__} for {self.device}>'



from __future__ import annotations

from typing import NamedTuple
from uuid import UUID

from pychromecast import get_chromecast_from_host, get_chromecasts, get_listed_chromecasts

from ..base import DEFAULT_DEVICE_NAME, DEFAULT_RETRY_WAIT, Device, NO_PORT, NO_STR, Seconds


class Host(NamedTuple):
  host: str
  port: int | None = NO_PORT
  uuid: str = NO_STR
  model_name: str = NO_STR
  friendly_name: str = DEFAULT_DEVICE_NAME


def get_device_via_host(
  host: str,
  name: str | None = DEFAULT_DEVICE_NAME,
  retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT,
) -> Device | None:
  name = name or DEFAULT_DEVICE_NAME
  info = Host(host, friendly_name=name)

  if device := get_chromecast_from_host(info, retry_wait=float(retry_wait)):
    device.wait()
    return device

  return None  # explicit


def get_devices(retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT) -> list[Device]:
  devices, service_browser = get_chromecasts(retry_wait=float(retry_wait))
  service_browser.stop_discovery()

  return devices


def get_listed_devices(
  name: str | None = None,
  uuid: UUID | str | None = None,
  retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT,
) -> list[Device]:
  devices, service_browser = get_listed_chromecasts(
    friendly_names=[name],
    uuids=[uuid],
    retry_wait=float(retry_wait),
  )
  service_browser.stop_discovery()

  return devices


def get_first(devices: list[Device]) -> Device | None:
  if not devices:
    return None

  first, *_ = devices
  first.wait()

  return first


def get_device_via_uuid(
  uuid: UUID | str | None = None,
  retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT,
) -> Device | None:
  devices = get_devices(retry_wait)

  if not uuid:
    return get_first(devices)

  uuid = UUID(uuid)

  for device in devices:
    if device.uuid == uuid:
      device.wait()

      return device

  if devices := get_listed_devices(uuid=uuid, retry_wait=retry_wait):
    return get_first(devices)

  return None


def get_device(
  name: str | None = None,
  retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT,
) -> Device | None:
  devices = get_devices(retry_wait)

  if not name:
    return get_first(devices)

  devices += get_listed_devices(name, retry_wait=retry_wait)

  name = name.casefold()

  for device in devices:
    if device.name.casefold() == name:
      device.wait()

      return device

  return None


def find_device(
  name: str | None = DEFAULT_DEVICE_NAME,
  host: str | None = None,
  uuid: UUID | str | None = None,
  retry_wait: Seconds | float | None = DEFAULT_RETRY_WAIT,
) -> Device | None:
  device: Device | None = None

  if host:
    device = get_device_via_host(host, name, retry_wait)

  if uuid and not device:
    device = get_device_via_uuid(uuid, retry_wait)

  if name and not device:
    device = get_device(name, retry_wait)

  no_identifiers = not (host or name or uuid)

  if no_identifiers:
    device = get_device(retry_wait=retry_wait)

  return device
